from abc import ABC, abstractmethod
import numpy as np
from PIL import ImageOps

import logging
log = logging.getLogger(__name__)

class AbstractPoseModel(ABC):
    """
        Abstract class for pose estimation models.
    """

    def __init__(self, tfengine):
                
        """Initialize posenet-base class with Tensorflow inference engine.
        :Parameters:
        ----------
        tfengine: Tensorflow inference engine.
        """

        self._tfengine = tfengine
        
        self._input_tensor_shape = self.get_input_tensor_shape()

        _, self._tensor_image_height, self._tensor_image_width, self._tensor_image_depth = \
            self.get_input_tensor_shape()

        self.confidence_threshold = self._tfengine.confidence_threshold
        log.debug(f"Initializing PoseEngine with confidence threshold \
            {self.confidence_threshold}")
        

    def get_input_tensor_shape(self):
        """Get the shape of the input tensor structure.
        Gets the shape required for the input tensor.
        For models trained for image classification / detection, the shape is
        always [1, height, width, channels].
        To be used as input for :func:`run_inference`,
        this tensor shape must be flattened into a 1-D array with size
        ``height * width * channels``. To instead get that 1-D array size, use
        :func:`required_input_array_size`.
        Returns:
        A 1-D array (:obj:`numpy.ndarray`) representing the required input
        tensor shape.
        """
        return self._tfengine.input_details[0]['shape']

 
    def tf_interpreter(self):
        return self._tfengine._tf_interpreter


    def thumbnail(self, image=None, desired_size=None):
        """Resizes original image as close as possible to desired size.
        Preserves aspect ratio of original image.
        Does not modify the original image.
        :Parameters:
        ----------
        image : PIL.Image
            Input Image for AI model detection.
        desired_size : (width, height)
            Size expected by the AI model.
        :Returns:
        -------
        PIL.Image
            Resized image fitting for the AI model input tensor.
        """
        assert image
        assert desired_size
        log.debug('input image size = %r', image.size)
        thumb = image.copy()
        w, h = desired_size
        try:
            # convert from numpy to native Python int type
            # that PIL expects
            if isinstance(w, np.generic):
                w = w.item()
                w = int(w)
                h = h.item()
                h = int(h)
            thumb.thumbnail((w, h))
        except Exception as e:
            msg = (f"Exception in "
                   f"PIL.image.thumbnail(desired_size={desired_size}):"
                   f"type(width)={type(w)}, type(height)={type(h)}"
                   f"\n{e}"
                   )
            log.exception(msg)
            raise RuntimeError(msg)
        log.debug('thmubnail image size = %r', thumb.size)
        return thumb


    def resize(self, image=None, desired_size=None):
        """Pad original image to exact size expected by input tensor.
        Preserve aspect ratio to avoid confusing the AI model with
        unnatural distortions. Pad the resulting image
        with solid black color pixels to fill the desired size.
        Do not modify the original image.
        :Parameters:
        ----------
        image : PIL.Image
            Input Image sized to fit an input tensor but without padding.
            Its possible that one size fits one tensor dimension exactly
            but the other size is smaller than
            the input tensor other dimension.
        desired_size : (width, height)
            Exact size expected by the AI model.
        :Returns:
        -------
        PIL.Image
            Resized image fitting exactly the AI model input tensor.
        """
        assert image
        assert desired_size
        log.debug('input image size = %r', image.size)
        thumb = image.copy()
        delta_w = desired_size[0] - thumb.size[0]
        delta_h = desired_size[1] - thumb.size[1]
        padding = (0, 0, delta_w, delta_h)
        new_im = ImageOps.expand(thumb, padding)
        log.debug('new image size = %r', new_im.size)
        assert new_im.size == desired_size
        return new_im


    @abstractmethod
    def execute_model(self, img):
        '''
            Execute Pose Estimation Model.
        '''


from src.pipeline.pose_base import AbstractPoseModel
import numpy as np
import time

class Posenet_MobileNet(AbstractPoseModel):
    '''The class for pose estimation using Posenet Mobilenet implementation.'''

    def __init__(self, tfengine):
        super().__init__(tfengine)


    def sigmoid(self, x):
        return 1 / (1 + np.exp(-x))


    def parse_output(self, heatmap_data, offset_data):
        '''
            Parse Output of TFLite model and get keypoints with score.
        '''

        joint_num = heatmap_data.shape[-1]
        pose_kps = np.zeros((joint_num, 3), np.float32)

        for i in range(heatmap_data.shape[-1]):

            joint_heatmap = heatmap_data[..., i]
            max_val_pos = np.squeeze(
                np.argwhere(joint_heatmap == np.max(joint_heatmap)))
            remap_pos = np.array(max_val_pos/8*self._tensor_image_height,
                                 dtype=np.int32)
            pose_kps[i, 0] = int(remap_pos[0] + offset_data[max_val_pos[0],
                                 max_val_pos[1], i])
            pose_kps[i, 1] = int(remap_pos[1] + offset_data[max_val_pos[0],
                                 max_val_pos[1], i+joint_num])
            max_prob = np.max(joint_heatmap)
            
            pose_kps[i,2] = self.sigmoid(max_prob)
            

        return pose_kps


    def execute_model(self, img):
        ''' Run TFLite model.
        
        :Parameters:
        ----------
        img: PIL.Image
            Input Image for AI model detection.
        :Returns:
        -------
        kps:
            A list of Pose objects with keypoints and confidence scores
        template_image: PIL.Image
            Input resized image.
        thumbnail: PIL.Image
            Thumbnail input image
        _inference_time: float
            Model inference time in seconds
        '''

        _tensor_input_size = (self._tensor_image_width,
                              self._tensor_image_height)

        # thumbnail is a proportionately resized image
        thumbnail = self.thumbnail(image=img,
                                   desired_size=_tensor_input_size)

        # convert thumbnail into an image with the exact size
        # as the input tensor preserving proportions by padding with
        # a solid color as needed
        template_image = self.resize(image=thumbnail,
                                desired_size=_tensor_input_size)

        start_time = time.process_time()

        template_input = np.expand_dims(template_image.copy(), axis=0)
        floating_model = self._tfengine.input_details[0]['dtype'] == np.float32

        if floating_model:
            template_input = (np.float32(template_input) - 127.5) / 127.5

        self.tf_interpreter().\
            set_tensor(self._tfengine.input_details[0]['index'],
                       template_input)
        self.tf_interpreter().invoke()

        template_output_data = self.tf_interpreter().\
            get_tensor(self._tfengine.output_details[0]['index'])
        template_offset_data = self.tf_interpreter().\
            get_tensor(self._tfengine.output_details[1]['index'])

        template_heatmaps = np.squeeze(template_output_data)
        template_offsets = np.squeeze(template_offset_data)

        kps = self.parse_output(template_heatmaps, template_offsets)

        _inference_time = time.process_time() - start_time

        return kps, template_image, thumbnail, _inference_time


from src import DEFAULT_DATA_DIR
import logging
import time
from PIL import ImageDraw
from pathlib import Path


from src.pipeline.posenet_model import Posenet_MobileNet
from src.pipeline.movenet_model import Movenet

log = logging.getLogger(__name__)

KEYPOINTS = (
  'nose',
  'left eye',
  'right eye',
  'left ear',
  'right ear',
  'left shoulder',
  'right shoulder',
  'left elbow',
  'right elbow',
  'left wrist',
  'right wrist',
  'left hip',
  'right hip',
  'left knee',
  'right knee',
  'left ankle',
  'right ankle'
)


class Keypoint:
    __slots__ = ['k', 'yx', 'score']

    def __init__(self, k, yx, score=None):
        self.k = k
        self.yx = yx
        self.score = score

    def __repr__(self):
        return 'Keypoint(<{}>, {}, {})'.format(self.k, self.yx, self.score)


class Pose:
    __slots__ = ['keypoints', 'score']

    def __init__(self, keypoints, score=None):
        assert len(keypoints) == len(KEYPOINTS)
        self.keypoints = keypoints
        self.score = score

    def __repr__(self):
        return 'Pose({}, {})'.format(self.keypoints, self.score)


class PoseEngine():
    """Engine used for pose tasks."""
    def __init__(self, tfengine=None, model_name=None, context=None):
        """Creates a PoseEngine wrapper around an initialized tfengine.
        """

        assert tfengine is not None
        assert model_name is not None

        if model_name == 'movenet':
            self._model = Movenet(tfengine)
        elif model_name == 'mobilenet':
            self._model = Posenet_MobileNet(tfengine)
        
        if context:
            self._sys_data_dir = context.data_dir
        else:
            self._sys_data_dir = DEFAULT_DATA_DIR
        self._sys_data_dir = Path(self._sys_data_dir)

        self.confidence_threshold = self._model.confidence_threshold
        self._tensor_image_height = self._model._tensor_image_height
        self._tensor_image_width = self._model._tensor_image_width

        

    def draw_kps(self, kps, template_image):

        pil_im = template_image
        draw = ImageDraw.Draw(pil_im)
        
        leftShoulder = False
        rightShoulder = False
        
        scoreList = {'LShoulder_score':0,'RShoulder_score':0,'LHip_score':0,'RHip_score':0}
        
        for i in range(kps.shape[0]):
                                    
            x, y, r = int(round(kps[i, 1])), int(round(kps[i, 0])), 1

            if i == 5:
                leftShoulder = True
                leftShoulder_point = [x, y]
                scoreList['LShoulder_score'] = kps[i,-1]
                
            if i == 6:
                rightShoulder = True
                rightShoulder_point = [x, y]
                scoreList['RShoulder_score'] = kps[i,-1]

            leftUpPoint = (x-r, y-r)
            rightDownPoint = (x+r, y+r)
            twoPointList = [leftUpPoint, rightDownPoint]
            draw.ellipse(twoPointList, fill=(0, 255, 0, 255))

            if i == 11 and leftShoulder:
                leftHip_point = [x, y]
                scoreList['LHip_score'] = kps[i,-1]
                draw.line((leftShoulder_point[0],leftShoulder_point[1], leftHip_point[0],leftHip_point[1]), fill='green', width=3)

            if i == 12 and rightShoulder:
                rightHip_point = [x, y]
                scoreList['RHip_score'] = kps[i,-1]
                draw.line((rightShoulder_point[0],rightShoulder_point[1], rightHip_point[0],rightHip_point[1]), fill='green', width=3)
                    
        return pil_im, scoreList


    def get_result(self, img):

        kps, template_image, thumbnail, _inference_time = self._model.execute_model(img)
        output_img, scoreList = self.draw_kps(kps, template_image)

        return thumbnail, output_img, scoreList, _inference_time


    def detect_poses(self, img):
        """
        Detects poses in a given image.
        :Parameters:
        ----------
        img : PIL.Image
            Input Image for AI model detection.
        :Returns:
        -------
        poses:
            A list of Pose objects with keypoints and confidence scores
        PIL.Image
            Resized image fitting the AI model input tensor.
        """

        kps, template_image, thumbnail, _ = self._model.execute_model(img)
        poses = []

        keypoint_dict = {}
        cnt = 0

        keypoint_count = kps.shape[0]
        for point_i in range(keypoint_count):
            x, y = kps[point_i, 1], kps[point_i, 0]
            prob = kps[point_i, 2]

            if prob > self.confidence_threshold and \
                0 < y < self._tensor_image_height and \
                0 < x < self._tensor_image_width:

                cnt += 1
                if log.getEffectiveLevel() <= logging.DEBUG:
                    # development mode
                    # draw on image and save it for debugging
                    draw = ImageDraw.Draw(template_image)
                    draw.line(((0, 0), (x, y)), fill='blue')
            else:
                prob = 0

            keypoint = Keypoint(KEYPOINTS[point_i], [x, y], prob)
            keypoint_dict[KEYPOINTS[point_i]] = keypoint

        # overall pose score is calculated as the average of all
        # individual keypoint scores
        pose_score = cnt/keypoint_count
        log.debug(f"Overall pose score (keypoint score average): {pose_score}")
        poses.append(Pose(keypoint_dict, pose_score))
        if cnt > 0 and log.getEffectiveLevel() <= logging.DEBUG:
            # development mode
            # save template_image for debugging
            timestr = int(time.monotonic()*1000)
            log.debug(f"Detected a pose with {cnt} keypoints that score over \
                the minimum confidence threshold of \
                {self.confidence_threshold}.")
            debug_image_file_name = \
                f'tmp-pose-detect-image-time-{timestr}-keypoints-{cnt}.jpg'
            # template_image.save(
            #                     Path(self._sys_data_dir,
            #                          debug_image_file_name),
            #                     format='JPEG')
            log.debug(f"Debug image saved: {debug_image_file_name}")
        return poses, thumbnail, pose_score



from src.pipeline.pose_base import AbstractPoseModel
import numpy as np
import time


class Movenet(AbstractPoseModel):
    '''The class for pose estimation using Movenet implementation.'''

    def __init__(self, tfengine):
        super().__init__(tfengine)


    def parse_output(self, keypoints_with_scores, height, width):
        '''
            Parse Output of TFLite model and get keypoints with score.
        '''
    
        keypoints_all = []
        num_instances, _, _, _ = keypoints_with_scores.shape
        
        for idx in range(num_instances):
            
            kpts_y = keypoints_with_scores[0, idx, :, 1]
            kpts_x = keypoints_with_scores[0, idx, :, 0]
            
            kpts_scores = keypoints_with_scores[0, idx, :, 2]
            
            kpts_absolute_xy = np.stack([width * np.array(kpts_x), height * np.array(kpts_y), kpts_scores], axis=-1)
            keypoints_all.append(kpts_absolute_xy)
            
        if keypoints_all:
            keypoints_xy = np.concatenate(keypoints_all, axis=0)
        else:
            keypoints_xy = np.zeros((0, 17, 2))

        return keypoints_xy


    def execute_model(self, img):
        ''' Run TFLite model.
        
        :Parameters:
        ----------
        img: PIL.Image
            Input Image for AI model detection.
        :Returns:
        -------
        kps:
            A list of Pose objects with keypoints and confidence scores
        template_image: PIL.Image
            Input resized image.
        thumbnail: PIL.Image
            Thumbnail input image
        _inference_time: float
            Model inference time in seconds
        '''
        
        _tensor_input_size = (self._tensor_image_width,
                              self._tensor_image_height)

        # thumbnail is a proportionately resized image
        thumbnail = self.thumbnail(image=img,
                                               desired_size=_tensor_input_size)
        # convert thumbnail into an image with the exact size
        # as the input tensor preserving proportions by padding with
        # a solid color as needed
        template_image = self.resize(image=thumbnail,
                                desired_size=_tensor_input_size)

        start_time = time.process_time()

        template_input = np.expand_dims(template_image.copy(), axis=0)
        floating_model = self._tfengine.input_details[0]['dtype'] == np.float32

        if floating_model:
            template_input = template_input.astype(np.float32)
            
        self.tf_interpreter().\
            set_tensor(self._tfengine.input_details[0]['index'],
                       template_input)
        self.tf_interpreter().invoke()

        keypoints_with_scores = self.tf_interpreter().get_tensor(self._tfengine.output_details[0]['index'])
        kps = self.parse_output(keypoints_with_scores, self._tensor_image_height, self._tensor_image_width)


        _inference_time = time.process_time() - start_time

        return kps, template_image, thumbnail, _inference_time


"""Tensorflow inference engine wrapper."""
import logging
import os
import numpy as np
from tflite_runtime.interpreter import Interpreter
from tflite_runtime.interpreter import load_delegate

log = logging.getLogger(__name__)


def _get_edgetpu_interpreter(model=None):  # pragma: no cover
    # Note: Looking for ideas how to test Coral EdgeTPU dependent code
    # in a cloud CI environment such as Travis CI and Github
    tf_interpreter = None
    if model:
        try:
            edgetpu_delegate = load_delegate('libedgetpu.so.1.0')
            assert edgetpu_delegate
            tf_interpreter = Interpreter(
                model_path=model,
                experimental_delegates=[edgetpu_delegate]
                )
            log.debug('EdgeTPU available. Will use EdgeTPU model.')
        except Exception as e:
            log.debug('EdgeTPU init error: %r', e)
            # log.debug(stacktrace())
    return tf_interpreter


class TFInferenceEngine:
    """Thin wrapper around TFLite Interpreter.

    The official TFLite API is moving fast and still changes frequently.
    This class intends to abstract out underlying TF changes to some extend.

    It dynamically detects if EdgeTPU is available and uses it.
    Otherwise falls back to TFLite Runtime.
    """

    def __init__(self,
                 model=None,
                 labels=None,
                 confidence_threshold=0.8,
                 **kwargs
                 ):
        """Create an instance of Tensorflow inference engine.

        :Parameters:
        ----------
        model: dict
            {
                'tflite': path,
                'edgetpu': path,
            }
            Where path is of type string and points to the
            location of frozen graph file (AI model).
        labels : string
            Location of file with model labels.
        confidence_threshold : float
            Inference confidence threshold.
        
        """
        assert model
        assert model['tflite'], 'TFLite AI model path required.'
        model_tflite = model['tflite']
        assert os.path.isfile(model_tflite), \
            'TFLite AI model file does not exist: {}' \
            .format(model_tflite)
        self._model_tflite_path = model_tflite
        model_edgetpu = model.get('edgetpu', None)
        if model_edgetpu:
            assert os.path.isfile(model_edgetpu), \
                'EdgeTPU AI model file does not exist: {}' \
                .format(model_edgetpu)
        self._model_edgetpu_path = model_edgetpu
        assert labels, 'AI model labels path required.'
        assert os.path.isfile(labels), \
            'AI model labels file does not exist: {}' \
            .format(labels)
        self._model_labels_path = labels
        self._confidence_threshold = confidence_threshold
        # log.info('Loading AI model:\n'
        #           'TFLite graph: %r\n'
        #           'EdgeTPU graph: %r\n'
        #           'Labels %r.'
        #           'Condidence threshod: %.0f%%'
        #           'top-k: %d',
        #           model_tflite,
        #           model_edgetpu,
        #           labels,
        #           confidence_threshold*100)
        # EdgeTPU is not available in testing and other environments
        # load dynamically as needed
#        edgetpu_class = 'DetectionEngine'
#        module_object = import_module('edgetpu.detection.engine',
#                                      packaage=edgetpu_class)
#        target_class = getattr(module_object, edgetpu_class)
        self._tf_interpreter = _get_edgetpu_interpreter(model=model_edgetpu)
        if not self._tf_interpreter:
            log.debug('EdgeTPU not available. Will use TFLite CPU runtime.')
            self._tf_interpreter = Interpreter(model_path=model_tflite)
        assert self._tf_interpreter
        self._tf_interpreter.allocate_tensors()
        # check the type of the input tensor
        self._tf_input_details = self._tf_interpreter.get_input_details()
        self._tf_output_details = self._tf_interpreter.get_output_details()
        self._tf_is_quantized_model = \
            self.input_details[0]['dtype'] != np.float32

    @property
    def input_details(self):
        return self._tf_input_details

    @property
    def output_details(self):
        return self._tf_output_details

    @property
    def is_quantized(self):
        return self._tf_is_quantized_model

    @property
    def confidence_threshold(self):
        """
        Inference confidence threshold.

        :Returns:
        -------
        float
            Confidence threshold for inference results.
            Only results at or above
            this threshold should be returned by each engine inference.

        """
        return self._confidence_threshold

    
    def infer(self):
        """Invoke model inference on current input tensor."""
        return self._tf_interpreter.invoke()

    def set_tensor(self, index=None, tensor_data=None):
        """Set tensor data at given reference index."""
        assert isinstance(index, int)
        self._tf_interpreter.set_tensor(index, tensor_data)

    def get_tensor(self, index=None):
        """Return tensor data at given reference index."""
        assert isinstance(index, int)
        return self._tf_interpreter.get_tensor(index)



"""Fall detection pipe element."""
from .inference import TFInferenceEngine
from src.pipeline.pose_engine import PoseEngine
from src import DEFAULT_DATA_DIR
import logging
import math
import time
from PIL import Image, ImageDraw
from pathlib import Path

log = logging.getLogger(__name__)


class FallDetector():

    """Detects falls comparing two images spaced about 1-2 seconds apart."""
    def __init__(self,
                 model=None,
                 labels=None,
                 confidence_threshold=0.15,
                 model_name=None,
                 **kwargs
                 ):
        """Initialize detector with config parameters.
        :Parameters:
        ----------
        model: dict
        {
            'tflite': 
                'ai_models/posenet_mobilenet_v1_100_257x257_multi_kpt_stripped.tflite'
            'edgetpu': 
                'ai_models/posenet_mobilenet_v1_075_721_1281_quant_decoder_edgetpu.tflite'
        }
        """
        
        self._tfengine = TFInferenceEngine(
                        model=model,
                        labels=labels,
                        confidence_threshold=confidence_threshold)
        self.model_name = model_name

        self._sys_data_dir = DEFAULT_DATA_DIR
        self._sys_data_dir = Path(self._sys_data_dir)

        # previous pose detection information for frame at time t-1 and t-2 \
        # to compare pose changes against
        self._prev_data = [None] * 2

        # Data of previous frames lookup constants
        self.POSE_VAL = '_prev_pose_dix'
        self.TIMESTAMP = '_prev_time'
        self.THUMBNAIL = '_prev_thumbnail'
        self.LEFT_ANGLE_WITH_YAXIS = '_prev_left_angle_with_yaxis'
        self.RIGHT_ANGLE_WITH_YAXIS = '_prev_right_angle_with_yaxis'
        self.BODY_VECTOR_SCORE = '_prev_body_vector_score'

        _dix = {self.POSE_VAL: [],
                self.TIMESTAMP: time.monotonic(),
                self.THUMBNAIL: None,
                self.LEFT_ANGLE_WITH_YAXIS: None,
                self.RIGHT_ANGLE_WITH_YAXIS: None,
                self.BODY_VECTOR_SCORE: 0
                }

        # self._prev_data[0] : store data of frame at t-2
        # self._prev_data[1] : store data of frame at t-1
        self._prev_data[0] = self._prev_data[1] = _dix

        self._pose_engine = PoseEngine(self._tfengine, self.model_name)
        self._fall_factor = 60
        self.confidence_threshold = confidence_threshold
        log.debug(f"Initializing FallDetector with conficence threshold: \
                  {self.confidence_threshold}")

        # Require a minimum amount of time between two video frames in seconds.
        # Otherwise on high performing hard, the poses could be too close to
        # each other and have negligible difference
        # for fall detection purpose.
        self.min_time_between_frames = 1
        # Require the time distance between two video frames not to exceed
        # a certain limit in seconds.
        # Otherwise there could be data noise which could lead
        # false positive detections.
        self.max_time_between_frames = 10

        # keypoint lookup constants
        self.LEFT_SHOULDER = 'left shoulder'
        self.LEFT_HIP = 'left hip'
        self.RIGHT_SHOULDER = 'right shoulder'
        self.RIGHT_HIP = 'right hip'

        self.fall_detect_corr = [self.LEFT_SHOULDER, self.LEFT_HIP,
                                 self.RIGHT_SHOULDER, self.RIGHT_HIP]

    def process_sample(self, **sample):
        """Detect objects in sample image."""
        log.debug("%s received new sample", self.__class__.__name__)
        if not sample:
            # pass through empty samples to next element
            yield None
        else:
            try:
                image = sample['image']
                inference_result, thumbnail = self.fall_detect(image=image)
                inference_result = self.convert_inference_result(
                                        inference_result)
                inf_meta = {
                    'display': 'Fall Detection',
                }
                # pass on the results to the next connected pipe element
                processed_sample = {
                    'image': image,
                    'thumbnail': thumbnail,
                    'inference_result': inference_result,
                    'inference_meta': inf_meta
                    }
                yield processed_sample
            except Exception as e:
                log.exception('Error "%s" while processing sample. '
                              'Dropping sample: %s',
                              str(e),
                              str(sample))

    def calculate_angle(self, p):
        '''
            Calculate angle b/w two lines such as
            left shoulder-hip with prev frame's left shoulder-hip or
            right shoulder-hip with prev frame's right shoulder-hip or
            shoulder-hip line with vertical axis
        '''

        x1, y1 = p[0][0]
        x2, y2 = p[0][1]

        x3, y3 = p[1][0]
        x4, y4 = p[1][1]

        theta1 = math.degrees(math.atan2(y2-y1, x1-x2))
        theta2 = math.degrees(math.atan2(y4-y3, x3-x4))

        angle = abs(theta1-theta2)
        return angle

    def is_body_line_motion_downward(self, left_angle_with_yaxis,
                                     rigth_angle_with_yaxis, inx):

        test = False

        l_angle = left_angle_with_yaxis and \
            self._prev_data[inx][self.LEFT_ANGLE_WITH_YAXIS]  \
            and left_angle_with_yaxis > \
            self._prev_data[inx][self.LEFT_ANGLE_WITH_YAXIS]
        r_angle = rigth_angle_with_yaxis and \
            self._prev_data[inx][self.RIGHT_ANGLE_WITH_YAXIS] \
            and rigth_angle_with_yaxis > \
            self._prev_data[inx][self.RIGHT_ANGLE_WITH_YAXIS]

        if l_angle or r_angle:
            test = True

        return test

    def find_keypoints(self, image):

        # this score value should be related to the configuration \
        # confidence_threshold parameter
        min_score = self.confidence_threshold
        rotations = [Image.ROTATE_270, Image.ROTATE_90]
        angle = 0
        pose = None
        poses, thumbnail, _ = self._pose_engine.detect_poses(image)
        width, height = thumbnail.size
        # if no pose detected with high confidence,
        # try rotating the image +/- 90' to find a fallen person
        # currently only looking at pose[0] because we are focused \
        # on a lone person falls
        # while (not poses or poses[0].score < min_score) and rotations:
        spinal_vector_score, pose_dix = self.estimate_spinal_vector_score(
                                        poses[0])
        while spinal_vector_score < min_score and rotations:
            angle = rotations.pop()
            transposed = image.transpose(angle)
            # we are interested in the poses but not the rotated thumbnail
            poses, _, _ = self._pose_engine.detect_poses(transposed)
            spinal_vector_score, pose_dix = self.estimate_spinal_vector_score(
                                    poses[0])

        if poses and poses[0]:
            pose = poses[0]

        # lets check if we found a good pose candidate

        if (pose and spinal_vector_score >= min_score):
            # if the image was rotated, we need to rotate back to the original\
            # image coordinates
            # before comparing with poses in other frames.
            if angle == Image.ROTATE_90:
                # ROTATE_90 rotates 90' counter clockwise \
                # from ^ to < orientation.
                for _, keypoint in pose.keypoints.items():
                    # keypoint.yx[0] is the x coordinate in an image
                    # keypoint.yx[0] is the y coordinate in an image, \
                    # with 0,0 in the upper left corner (not lower left).
                    tmp_swap = keypoint.yx[0]
                    keypoint.yx[0] = width-keypoint.yx[1]
                    keypoint.yx[1] = tmp_swap
            elif angle == Image.ROTATE_270:
                # ROTATE_270 rotates 90' clockwise from ^ to > orientation.
                for _, keypoint in pose.keypoints.items():
                    tmp_swap = keypoint.yx[0]
                    keypoint.yx[0] = keypoint.yx[1]
                    keypoint.yx[1] = height-tmp_swap
            # we could not detexct a pose with sufficient confidence
            log.info(f"""A pose detected with
                    spinal_vector_score={spinal_vector_score} >= {min_score}
                    confidence threshold.
                    Pose keypoints: {pose_dix}"
                """)
        else:
            pose = None

        return pose, thumbnail, spinal_vector_score, pose_dix

    def find_changes_in_angle(self, pose_dix, inx):
        '''
            Find the changes in angle for shoulder-hip lines
            b/w current and previpus frame.
        '''

        prev_leftLine_corr_exist = all(e in self._prev_data[inx][self.POSE_VAL]
                                       for e in [self.LEFT_SHOULDER, self.LEFT_HIP])
        curr_leftLine_corr_exist = all(e in pose_dix for e in [self.LEFT_SHOULDER,self.LEFT_HIP])

        prev_rightLine_corr_exist = all(e in self._prev_data[inx][self.POSE_VAL] for e in [self.RIGHT_SHOULDER, self.RIGHT_HIP])
        curr_rightLine_corr_exist = all(e in pose_dix for e in [self.RIGHT_SHOULDER, self.RIGHT_HIP])

        left_angle = right_angle = 0

        if prev_leftLine_corr_exist and curr_leftLine_corr_exist:
            temp_left_vector = [[self._prev_data[inx][self.POSE_VAL][self.LEFT_SHOULDER],
                                self._prev_data[inx][self.POSE_VAL][self.LEFT_HIP]],
                                [pose_dix[self.LEFT_SHOULDER], pose_dix[self.LEFT_HIP]]]
            left_angle = self.calculate_angle(temp_left_vector)
            log.debug("Left shoulder-hip angle: %r", left_angle)

        if prev_rightLine_corr_exist and curr_rightLine_corr_exist:
            temp_right_vector = [[self._prev_data[inx][self.POSE_VAL][self.RIGHT_SHOULDER],
                                 self._prev_data[inx][self.POSE_VAL][self.RIGHT_HIP]],
                                 [pose_dix[self.RIGHT_SHOULDER], pose_dix[self.RIGHT_HIP]]]
            right_angle = self.calculate_angle(temp_right_vector)
            log.debug("Right shoulder-hip angle: %r", right_angle)

        angle_change = max(left_angle, right_angle)
        return angle_change

    def assign_prev_records(self, pose_dix, left_angle_with_yaxis,
                            rigth_angle_with_yaxis, now, thumbnail,
                            current_body_vector_score):

        curr_data = {self.POSE_VAL: pose_dix,
                     self.TIMESTAMP: now,
                     self.THUMBNAIL: thumbnail,
                     self.LEFT_ANGLE_WITH_YAXIS: left_angle_with_yaxis,
                     self.RIGHT_ANGLE_WITH_YAXIS: rigth_angle_with_yaxis,
                     self.BODY_VECTOR_SCORE: current_body_vector_score}

        self._prev_data[-2] = self._prev_data[-1]
        self._prev_data[-1] = curr_data

    def draw_lines(self, thumbnail, pose_dix, score):
        """Draw body lines if available. Return number of lines drawn."""
        # save an image with drawn lines for debugging
        draw = ImageDraw.Draw(thumbnail)
        path = None
        body_lines_drawn = 0

        if pose_dix is None:
            return body_lines_drawn

        if pose_dix.keys() >= {self.LEFT_SHOULDER, self.LEFT_HIP}:
            body_line = [tuple(pose_dix[self.LEFT_SHOULDER]),
                         tuple(pose_dix[self.LEFT_HIP])]
            draw.line(body_line, fill='red')
            body_lines_drawn += 1

        if pose_dix.keys() >= {self.RIGHT_SHOULDER, self.RIGHT_HIP}:
            body_line = [tuple(pose_dix[self.RIGHT_SHOULDER]),
                         tuple(pose_dix[self.RIGHT_HIP])]
            draw.line(body_line, fill='red')
            body_lines_drawn += 1

        # save a thumbnail for debugging
        timestr = int(time.monotonic()*1000)
        debug_image_file_name = \
            f'tmp-fall-detect-thumbnail-{timestr}-score-{score}.jpg'
        # thumbnail.save(
        #                Path(self._sys_data_dir, debug_image_file_name),
        #                format='JPEG')
        print(Path(self._sys_data_dir, debug_image_file_name))
        return body_lines_drawn

    def get_line_angles_with_yaxis(self, pose_dix):
        '''
            Find the angle b/w shoulder-hip line with yaxis.
        '''
        y_axis_corr = [[0, 0], [0, self._pose_engine._tensor_image_height]]

        leftLine_corr_exist = all(e in pose_dix for e in [self.LEFT_SHOULDER, self.LEFT_HIP])
        rightLine_corr_exist = all(e in pose_dix for e in [self.RIGHT_SHOULDER, self.RIGHT_HIP])

        l_angle = r_angle = 0

        if leftLine_corr_exist:
            l_angle = self.calculate_angle([y_axis_corr,
                                           [pose_dix[self.LEFT_SHOULDER],
                                            pose_dix[self.LEFT_HIP]]])

        if rightLine_corr_exist:
            r_angle = self.calculate_angle([y_axis_corr,
                                           [pose_dix[self.RIGHT_SHOULDER],
                                            pose_dix[self.RIGHT_HIP]]])

        return (l_angle, r_angle)

    def estimate_spinal_vector_score(self, pose):
        pose_dix = {}
        is_leftVector = is_rightVector = False

        # Calculate leftVectorScore & rightVectorScore
        leftVectorScore = min(pose.keypoints[self.LEFT_SHOULDER].score,
                              pose.keypoints[self.LEFT_HIP].score)
        rightVectorScore = min(pose.keypoints[self.RIGHT_SHOULDER].score,
                               pose.keypoints[self.RIGHT_HIP].score)

        if leftVectorScore > self.confidence_threshold:
            is_leftVector = True
            pose_dix[self.LEFT_SHOULDER] = \
                pose.keypoints[self.LEFT_SHOULDER].yx
            pose_dix[self.LEFT_HIP] = pose.keypoints[self.LEFT_HIP].yx

        if rightVectorScore > self.confidence_threshold:
            is_rightVector = True
            pose_dix[self.RIGHT_SHOULDER] = \
                pose.keypoints[self.RIGHT_SHOULDER].yx
            pose_dix[self.RIGHT_HIP] = pose.keypoints[self.RIGHT_HIP].yx

        def find_spinalLine():
            left_spinal_x1 = (pose_dix[self.LEFT_SHOULDER][0] +
                              pose_dix[self.RIGHT_SHOULDER][0]) / 2
            left_spinal_y1 = (pose_dix[self.LEFT_SHOULDER][1] +
                              pose_dix[self.RIGHT_SHOULDER][1]) / 2

            right_spinal_x1 = (pose_dix[self.LEFT_HIP][0] +
                               pose_dix[self.RIGHT_HIP][0]) / 2
            right_spinal_y1 = (pose_dix[self.LEFT_HIP][1] +
                               pose_dix[self.RIGHT_HIP][1]) / 2

            return (left_spinal_x1, left_spinal_y1), \
                   (right_spinal_x1, right_spinal_y1)

        if is_leftVector and is_rightVector:
            spinalVectorEstimate = find_spinalLine()
            spinalVectorScore = (leftVectorScore + rightVectorScore) / 2.0
        elif is_leftVector:
            spinalVectorEstimate = pose_dix[self.LEFT_SHOULDER], \
                                   pose_dix[self.LEFT_HIP]
            # 10% score penalty in conficence as only \
            # left shoulder-hip line is detected
            spinalVectorScore = leftVectorScore * 0.9
        elif is_rightVector:
            spinalVectorEstimate = pose_dix[self.RIGHT_SHOULDER], \
                                   pose_dix[self.RIGHT_HIP]
            # 10% score penalty in conficence as only \
            # right shoulder-hip line is detected
            spinalVectorScore = rightVectorScore * 0.9
        else:
            spinalVectorScore = 0

        log.debug(f"Estimated spinal vector score: {spinalVectorScore}")
        return spinalVectorScore, pose_dix

    def fall_detect(self, image=None):
        assert image
        log.debug("Calling TF engine for inference")
        start_time = time.monotonic()

        now = time.monotonic()
        lapse = now - self._prev_data[-1][self.TIMESTAMP]

        if self._prev_data[-1][self.POSE_VAL] \
           and lapse < self.min_time_between_frames:
            log.debug("Received an image frame too soon after the previous \
                frame. Only %.2f ms apart.\
                Minimum %.2f ms distance required for fall detection.",
                lapse, self.min_time_between_frames)
            inference_result = None
            thumbnail = self._prev_data[-1][self.THUMBNAIL]
        else:
            # Detection using tensorflow posenet module
            pose, thumbnail, spinal_vector_score, pose_dix = \
                        self.find_keypoints(image)

            inference_result = None
            if not pose:
                log.debug(f"No pose detected or detection score does not meet \
                    confidence threshold of {self.confidence_threshold}.")
            else:
                inference_result = []

                current_body_vector_score = spinal_vector_score

                # Find line angle with vertcal axis
                left_angle_with_yaxis, rigth_angle_with_yaxis = \
                    self.get_line_angles_with_yaxis(pose_dix)

                # save an image with drawn lines for debugging
                if log.getEffectiveLevel() <= logging.DEBUG:
                    # development mode
                    self.draw_lines(thumbnail, pose_dix, spinal_vector_score)

                for t in [-1, -2]:
                    lapse = now - self._prev_data[t][self.TIMESTAMP]

                    if not self._prev_data[t][self.POSE_VAL] or \
                       lapse > self.max_time_between_frames:
                        log.debug("No recent pose to compare to. Will save \
                            this frame pose for subsequent comparison.")
                    elif not self.is_body_line_motion_downward(
                                                        left_angle_with_yaxis,
                                                        rigth_angle_with_yaxis,
                                                        inx=t):
                        log.debug("The body-line angle with vertical axis is \
                                    decreasing from the previous frame. \
                                    Not likely to be a fall.")
                    else:
                        leaning_angle = self.find_changes_in_angle(pose_dix,
                                                                   inx=t)

                        # Get leaning_probability by comparing leaning_angle
                        # with fall_factor probability.
                        leaning_probability = 1 \
                            if leaning_angle > self._fall_factor else 0

                        # Calculate fall score using average of current and \
                        # previous frame's body vector score with \
                        # leaning_probability
                        fall_score = leaning_probability * \
                            (self._prev_data[t][self.BODY_VECTOR_SCORE] +
                             current_body_vector_score) / 2

                        if fall_score >= self.confidence_threshold:
                            inference_result.append(('FALL', fall_score,
                                                     leaning_angle, pose_dix))
                            log.info("Fall detected: %r", inference_result)
                            break
                        else:
                            log.debug(f"No fall detected due to low \
                            confidence score:  \
                            {fall_score} < {self.confidence_threshold} \
                            min threshold.Inference result: {inference_result}")

                log.debug("Saving pose for subsequent comparison.")
                self.assign_prev_records(pose_dix, left_angle_with_yaxis,
                                         rigth_angle_with_yaxis, now,
                                         thumbnail,
                                         current_body_vector_score)

                # log.debug("Logging stats")

        # self.log_stats(start_time=start_time)
        log.debug("thumbnail: %r", thumbnail)
        return inference_result, thumbnail

    def convert_inference_result(self, inference_result):
        inf_json = []

        if inference_result:
            for inf in inference_result:
                label, confidence, leaning_angle, keypoint_corr = inf
                log.info('label: %s , confidence: %.0f, leaning_angle: %.0f, \
                         keypoint_corr: %s',
                         label,
                         confidence,
                         leaning_angle,
                         keypoint_corr)
                one_inf = {
                    'label': label,
                    'confidence': confidence,
                    'leaning_angle': leaning_angle,
                    'keypoint_corr': {
                        'left shoulder': keypoint_corr.get('left shoulder',
                                                           None),
                        'left hip': keypoint_corr.get('left hip', None),
                        'right shoulder': keypoint_corr.get('right shoulder',
                                                            None),
                        'right hip': keypoint_corr.get('right hip', None)
                    }
                }
                inf_json.append(one_inf)

        return inf_json
